@@ -3,6 +3,7 @@
 #
 
 from Symtab import BuiltinScope, StructOrUnionScope
+from Cython.Utils import UtilityCode
 from TypeSlots import Signature
 import PyrexTypes
 
@@ -30,6 +31,7 @@ builtin_function_table = [
     ('issubclass', "OO",   "b",     "PyObject_IsSubclass"),
     ('iter',       "O",    "O",     "PyObject_GetIter"),
     ('len',        "O",    "Z",     "PyObject_Length"),
+    ('locals',     "",     "O",     "__pyx_locals"),
     #('map',       "",     "",      ""),
     #('max',       "",     "",      ""),
     #('min',       "",     "",      ""),
@@ -82,43 +84,61 @@ builtin_function_table = [
 builtin_types_table = [
 
     ("type",    "PyType_Type",     []),
-#    ("str",     "PyBytes_Type",   []),
+
+    ("bool",    "PyBool_Type",     []),
+    ("int",     "PyInt_Type",      []),
+    ("long",    "PyLong_Type",     []),
+    ("float",   "PyFloat_Type",    []),
+    ("complex", "PyComplex_Type",  []),
+
+    ("bytes",   "PyBytes_Type",    []),
+    ("str",     "PyString_Type",   []),
     ("unicode", "PyUnicode_Type",  []),
-    ("file",    "PyFile_Type",     []),
-#    ("slice",   "PySlice_Type",    []),
-#    ("set",     "PySet_Type",      []),
-    ("frozenset", "PyFrozenSet_Type",   []),
 
     ("tuple",   "PyTuple_Type",    []),
-    
+
     ("list",    "PyList_Type",     [("append", "OO",   "i", "PyList_Append"),
                                     ("insert", "OiO",  "i", "PyList_Insert"),
                                     ("sort",   "O",    "i", "PyList_Sort"),
                                     ("reverse","O",    "i", "PyList_Reverse")]),
-                                    
+
     ("dict",    "PyDict_Type",     [("items", "O",   "O", "PyDict_Items"),
                                     ("keys",  "O",   "O", "PyDict_Keys"),
                                     ("values","O",   "O", "PyDict_Values")]),
+
+    ("slice",   "PySlice_Type",    []),
+    ("file",    "PyFile_Type",     []),
+
+    ("set",       "PySet_Type",    [("clear",   "O",  "i", "PySet_Clear"), 
+                                    ("discard", "OO", "i", "PySet_Discard"),
+                                    ("add",     "OO", "i", "PySet_Add"),
+                                    ("pop",     "O",  "O", "PySet_Pop")]),
+    ("frozenset", "PyFrozenSet_Type", []),
 ]
 
+
+        
 builtin_structs_table = [
     ('Py_buffer', 'Py_buffer',
      [("buf",        PyrexTypes.c_void_ptr_type),
+      ("obj",        PyrexTypes.py_object_type),
       ("len",        PyrexTypes.c_py_ssize_t_type),
+      ("itemsize",   PyrexTypes.c_py_ssize_t_type),
       ("readonly",   PyrexTypes.c_bint_type),
-      ("format",     PyrexTypes.c_char_ptr_type),
       ("ndim",       PyrexTypes.c_int_type),
+      ("format",     PyrexTypes.c_char_ptr_type),
       ("shape",      PyrexTypes.c_py_ssize_t_ptr_type),
       ("strides",    PyrexTypes.c_py_ssize_t_ptr_type),
       ("suboffsets", PyrexTypes.c_py_ssize_t_ptr_type),
-      ("itemsize",   PyrexTypes.c_py_ssize_t_type),
       ("internal",   PyrexTypes.c_void_ptr_type),
       ])
 ]
 
-getattr3_utility_code = ["""
+getattr3_utility_code = UtilityCode(
+proto = """
 static PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *); /*proto*/
-""","""
+""",
+impl = """
 static PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
     PyObject *r = PyObject_GetAttr(o, n);
     if (!r) {
@@ -132,20 +152,131 @@ static PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
 bad:
     return 0;
 }
-"""]
+""")
 
-intern_utility_code = ["""
+intern_utility_code = UtilityCode(
+proto = """
 #if PY_MAJOR_VERSION >= 3
 #  define __Pyx_InternFromString(s) PyUnicode_InternFromString(s)
 #else
 #  define __Pyx_InternFromString(s) PyString_InternFromString(s)
 #endif
-""","""
-"""]
+""")
+
+def put_py23_set_init_utility_code(code, pos):
+    code.putln("#if PY_VERSION_HEX < 0x02040000")
+    code.putln(code.error_goto_if_neg("__Pyx_Py23SetsImport()", pos))
+    code.putln("#endif")
+
+py23_set_utility_code = UtilityCode(
+proto = """
+#if PY_VERSION_HEX < 0x02050000
+#ifndef PyAnySet_CheckExact
+
+#define PyAnySet_CheckExact(ob) \\
+    ((ob)->ob_type == &PySet_Type || \\
+     (ob)->ob_type == &PyFrozenSet_Type)
+
+#define PySet_New(iterable) \\
+    PyObject_CallFunctionObjArgs((PyObject *)&PySet_Type, (iterable), NULL)
+
+#define Pyx_PyFrozenSet_New(iterable) \\
+    PyObject_CallFunctionObjArgs((PyObject *)&PyFrozenSet_Type, (iterable), NULL)
+
+#define PySet_Size(anyset) \\
+    PyObject_Size((anyset))
+
+#define PySet_Contains(anyset, key) \\
+    PySequence_Contains((anyset), (key))
+
+#define PySet_Pop(set) \\
+    PyObject_CallMethod(set, "pop", NULL)
+
+static INLINE int PySet_Clear(PyObject *set) {
+    PyObject *ret = PyObject_CallMethod(set, "clear", NULL);
+    if (!ret) return -1;
+    Py_DECREF(ret); return 0;
+}
+
+static INLINE int PySet_Discard(PyObject *set, PyObject *key) {
+    PyObject *ret = PyObject_CallMethod(set, "discard", "O", key);
+    if (!ret) return -1;
+    Py_DECREF(ret); return 0;
+}
+
+static INLINE int PySet_Add(PyObject *set, PyObject *key) {
+    PyObject *ret = PyObject_CallMethod(set, "add", "O", key);
+    if (!ret) return -1;
+    Py_DECREF(ret); return 0;
+}
+
+#endif /* PyAnySet_CheckExact (<= Py2.4) */
+
+#if PY_VERSION_HEX < 0x02040000
+#ifndef Py_SETOBJECT_H
+#define Py_SETOBJECT_H
+
+static PyTypeObject *__Pyx_PySet_Type = NULL;
+static PyTypeObject *__Pyx_PyFrozenSet_Type = NULL;
+
+#define PySet_Type (*__Pyx_PySet_Type)
+#define PyFrozenSet_Type (*__Pyx_PyFrozenSet_Type)
+
+#define PyAnySet_Check(ob) \\
+    (PyAnySet_CheckExact(ob) || \\
+     PyType_IsSubtype((ob)->ob_type, &PySet_Type) || \\
+     PyType_IsSubtype((ob)->ob_type, &PyFrozenSet_Type))
+
+#define PyFrozenSet_CheckExact(ob) ((ob)->ob_type == &PyFrozenSet_Type)
+
+static int __Pyx_Py23SetsImport(void) {
+    PyObject *sets=0, *Set=0, *ImmutableSet=0;
+
+    sets = PyImport_ImportModule("sets");
+    if (!sets) goto bad;
+    Set = PyObject_GetAttrString(sets, "Set");
+    if (!Set) goto bad;
+    ImmutableSet = PyObject_GetAttrString(sets, "ImmutableSet");
+    if (!ImmutableSet) goto bad;
+    Py_DECREF(sets);
+  
+    __Pyx_PySet_Type       = (PyTypeObject*) Set;
+    __Pyx_PyFrozenSet_Type = (PyTypeObject*) ImmutableSet;
+
+    /* FIXME: this should be done in dedicated module cleanup code */
+    /*
+    Py_DECREF(Set);
+    Py_DECREF(ImmutableSet);
+    */
+
+    return 0;
+
+ bad:
+    Py_XDECREF(sets);
+    Py_XDECREF(Set);
+    Py_XDECREF(ImmutableSet);
+    return -1;
+}
+
+#else
+static int __Pyx_Py23SetsImport(void) { return 0; }
+#endif /* !Py_SETOBJECT_H */
+#endif /* < Py2.4  */
+#endif /* < Py2.5  */
+""",
+init = put_py23_set_init_utility_code,
+cleanup = """
+#if PY_VERSION_HEX < 0x02040000
+Py_XDECREF(__Pyx_PySet_Type); __Pyx_PySet_Type = NULL;
+Py_XDECREF(__Pyx_PyFrozenSet_Type); __Pyx_PyFrozenSet_Type = NULL;
+#endif /* < Py2.4  */
+""")
 
 builtin_utility_code = {
-    'getattr3': getattr3_utility_code,
-    'intern'  : intern_utility_code,
+    'getattr3'  : getattr3_utility_code,
+    'intern'    : intern_utility_code,
+    'set'       : py23_set_utility_code,
+    'frozenset' : py23_set_utility_code,
 }
 
 builtin_scope = BuiltinScope()
@@ -162,7 +293,8 @@ def init_builtin_funcs():
 
 def init_builtin_types():
     for name, cname, funcs in builtin_types_table:
-        the_type = builtin_scope.declare_builtin_type(name, cname)
+        utility = builtin_utility_code.get(name)
+        the_type = builtin_scope.declare_builtin_type(name, cname, utility)
         for name, args, ret, cname in funcs:
             sig = Signature(args, ret)
             the_type.scope.declare_cfunction(name, sig.function_type(), None, cname)
@@ -180,9 +312,11 @@ def init_builtins():
     init_builtin_funcs()
     init_builtin_types()
     init_builtin_structs()
-    global list_type, tuple_type, dict_type
+    global list_type, tuple_type, dict_type, unicode_type, type_type
+    type_type  = builtin_scope.lookup('type').type
     list_type  = builtin_scope.lookup('list').type
     tuple_type = builtin_scope.lookup('tuple').type
     dict_type  = builtin_scope.lookup('dict').type
+    unicode_type = builtin_scope.lookup('unicode').type
 
 init_builtins()
