@@ -2,8 +2,10 @@
 #   Pyrex - Types
 #
 
-from Cython import Utils
+from Cython.Utils import UtilityCode
+import StringEncoding
 import Naming
+import copy
 
 class BaseType:
     #
@@ -27,6 +29,7 @@ class PyrexType(BaseType):
     #  is_extension_type     boolean     Is a Python extension type
     #  is_numeric            boolean     Is a C numeric type
     #  is_int                boolean     Is a C integer type
+    #  is_longlong           boolean     Is a long long or unsigned long long.
     #  is_float              boolean     Is a C floating point type
     #  is_void               boolean     Is the C void type
     #  is_array              boolean     Is a C array type
@@ -34,12 +37,14 @@ class PyrexType(BaseType):
     #  is_null_ptr           boolean     Is the type of NULL
     #  is_cfunction          boolean     Is a C function type
     #  is_struct_or_union    boolean     Is a C struct or union type
+    #  is_struct             boolean     Is a C struct type
     #  is_enum               boolean     Is a C enum type
     #  is_typedef            boolean     Is a typedef type
     #  is_string             boolean     Is a C char * type
     #  is_unicode            boolean     Is a UTF-8 encoded C char * type
     #  is_returncode         boolean     Is used only to signal exceptions
     #  is_error              boolean     Is the dummy error type
+    #  is_buffer             boolean     Is buffer access type
     #  has_attributes        boolean     Has C dot-selectable attributes
     #  default_value         string      Initial value
     #  parsetuple_format     string      Format char for PyArg_ParseTuple
@@ -75,6 +80,7 @@ class PyrexType(BaseType):
     is_builtin_type = 0
     is_numeric = 0
     is_int = 0
+    is_longlong = 0
     is_float = 0
     is_void = 0
     is_array = 0
@@ -82,12 +88,14 @@ class PyrexType(BaseType):
     is_null_ptr = 0
     is_cfunction = 0
     is_struct_or_union = 0
+    is_struct = 0
     is_enum = 0
     is_typedef = 0
     is_string = 0
     is_unicode = 0
     is_returncode = 0
     is_error = 0
+    is_buffer = 0
     has_attributes = 0
     default_value = ""
     parsetuple_format = ""
@@ -131,6 +139,9 @@ class PyrexType(BaseType):
         # a struct whose attributes are not defined, etc.
         return 1
 
+    def is_simple_buffer_dtype(self):
+        return (self.is_int or self.is_float or self.is_pyobject or
+                self.is_extension_type or self.is_ptr)
 
 class CTypedefType(BaseType):
     #
@@ -183,16 +194,52 @@ class CTypedefType(BaseType):
     def __getattr__(self, name):
         return getattr(self.typedef_base_type, name)
 
+class BufferType(BaseType):
+    #
+    #  Delegates most attribute
+    #  lookups to the base type. ANYTHING NOT DEFINED
+    #  HERE IS DELEGATED!
+    
+    # dtype            PyrexType
+    # ndim             int
+    # mode             str
+    # negative_indices bool
+    # cast             bool
+    # is_buffer        bool
+    # writable         bool
 
+    is_buffer = 1
+    writable = True
+    def __init__(self, base, dtype, ndim, mode, negative_indices, cast):
+        self.base = base
+        self.dtype = dtype
+        self.ndim = ndim
+        self.buffer_ptr_type = CPtrType(dtype)
+        self.mode = mode
+        self.negative_indices = negative_indices
+        self.cast = cast
+    
+    def as_argument_type(self):
+        return self
+
+    def __getattr__(self, name):
+        return getattr(self.base, name)
+
+    def __repr__(self):
+        return "<BufferType %r>" % self.base
+
+    
 class PyObjectType(PyrexType):
     #
     #  Base class for all Python object types (reference-counted).
     #
+    #  buffer_defaults  dict or None     Default options for bu
     
     is_pyobject = 1
     default_value = "0"
     parsetuple_format = "O"
     pymemberdef_typecode = "T_OBJECT"
+    buffer_defaults = None
     
     def __str__(self):
         return "Python object"
@@ -250,7 +297,23 @@ class BuiltinObjectType(PyObjectType):
         return type.is_pyobject and self.assignable_from(type)
         
     def type_test_code(self, arg):
-        return 'likely(Py%s_CheckExact(%s)) || (%s) == Py_None || (PyErr_Format(PyExc_TypeError, "Expected %s, got %%s", Py_TYPE(%s)->tp_name), 0)' % (self.name[0].upper() + self.name[1:], arg, arg, self.name, arg)
+        type_name = self.name
+        if type_name == 'str':
+            type_name = 'String'
+        elif type_name == 'set':
+            type_name = 'AnySet'
+        elif type_name == 'frozenset':
+            type_name = 'FrozenSet'
+        else:
+            type_name = type_name.capitalize()
+        return 'likely(Py%s_CheckExact(%s)) || (%s) == Py_None || (PyErr_Format(PyExc_TypeError, "Expected %s, got %%s", Py_TYPE(%s)->tp_name), 0)' % (type_name, arg, arg, self.name, arg)
+
+    def declaration_code(self, entity_code, 
+            for_display = 0, dll_linkage = None, pyrex = 0):
+        if pyrex or for_display:
+            return self.base_declaration_code(self.name, entity_code)
+        else:
+            return "%s *%s" % (public_decl("PyObject", dll_linkage), entity_code)
 
 
 class PyExtensionType(PyObjectType):
@@ -350,7 +413,10 @@ class CType(PyrexType):
     from_py_function = None
     exception_value = None
     exception_check = 1
-    
+
+    def create_convert_utility_code(self, env):
+        return True
+        
     def error_condition(self, result_code):
         conds = []
         if self.is_string:
@@ -419,6 +485,8 @@ class CNumericType(CType):
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
         base = public_decl(self.sign_and_name(), dll_linkage)
+        if for_display and self.is_longlong:
+            base = base.replace('PY_LONG_LONG', 'long long')
         return self.base_declaration_code(base,  entity_code)
 
 
@@ -502,14 +570,16 @@ class CULongType(CUIntType):
     from_py_function = "PyInt_AsUnsignedLongMask"
 
 
-class CLongLongType(CUIntType):
+class CLongLongType(CIntType):
 
+    is_longlong = 1
     to_py_function = "PyLong_FromLongLong"
     from_py_function = "__pyx_PyInt_AsLongLong"
 
 
 class CULongLongType(CUIntType):
 
+    is_longlong = 1
     to_py_function = "PyLong_FromUnsignedLongLong"
     from_py_function = "__pyx_PyInt_AsUnsignedLongLong"
 
@@ -679,7 +749,7 @@ class CFuncType(CType):
             return 1
         if not other_type.is_cfunction:
             return 0
-        if not self.is_overridable and other_type.is_overridable:
+        if self.is_overridable != other_type.is_overridable:
             return 0
         nargs = len(self.args)
         if nargs != len(other_type.args):
@@ -793,6 +863,8 @@ class CFuncType(CType):
         for arg in self.args[:len(self.args)-self.optional_arg_count]:
             arg_decl_list.append(
                 arg.type.declaration_code("", for_display, pyrex = pyrex))
+        if self.is_overridable:
+            arg_decl_list.append("int %s" % Naming.skip_dispatch_cname)
         if self.optional_arg_count:
             arg_decl_list.append(self.op_arg_struct.declaration_code(Naming.optional_args_cname))
         if self.has_varargs:
@@ -835,9 +907,12 @@ class CFuncTypeArg:
     #  type       PyrexType
     #  pos        source file position
     
-    def __init__(self, name, type, pos):
+    def __init__(self, name, type, pos, cname=None):
         self.name = name
-        self.cname = Naming.var_prefix + name
+        if cname is not None:
+            self.cname = cname
+        else:
+            self.cname = Naming.var_prefix + name
         self.type = type
         self.pos = pos
         self.not_none = False
@@ -866,6 +941,50 @@ class CStructOrUnionType(CType):
         self.kind = kind
         self.scope = scope
         self.typedef_flag = typedef_flag
+        self.is_struct = kind == 'struct'
+        if self.is_struct:
+            self.to_py_function = "%s_to_py_%s" % (Naming.convert_func_prefix, self.cname)
+        self.exception_check = True
+        self._convert_code = None
+        
+    def create_convert_utility_code(self, env):
+        if env.outer_scope is None:
+            return False
+        if self._convert_code is None:
+            import Code
+            code = Code.CCodeWriter()
+            header = "static PyObject* %s(%s)" % (self.to_py_function, self.declaration_code('s'))
+            code.putln("%s {" % header)
+            code.putln("PyObject* res;")
+            code.putln("PyObject* member;")
+            code.putln("res = PyDict_New(); if (res == NULL) return NULL;")
+            for member in self.scope.var_entries:
+                if member.type.to_py_function and member.type.create_convert_utility_code(env):
+                    interned_name = env.get_string_const(member.name, identifier=True)
+                    env.add_py_string(interned_name)
+                    code.putln("member = %s(s.%s); if (member == NULL) goto bad;" % (
+                                                member.type.to_py_function, member.cname))
+                    code.putln("if (PyDict_SetItem(res, %s, member) < 0) goto bad;" % interned_name.pystring_cname)
+                    code.putln("Py_DECREF(member);")
+                else:
+                    self.to_py_function = None
+                    return False
+            code.putln("return res;")
+            code.putln("bad:")
+            code.putln("Py_XDECREF(member);")
+            code.putln("Py_DECREF(res);")
+            code.putln("return NULL;")
+            code.putln("}")
+            proto = header + ";"
+            # This is a bit of a hack, we need a forward declaration
+            # due to the way things are ordered in the module...
+            entry = env.lookup(self.name)
+            if entry.visibility != 'extern':
+                proto = self.declaration_code('') + ';\n' + proto
+            self._convert_code = UtilityCode(proto=proto, impl=code.buffer.getvalue())
+        
+        env.use_utility_code(self._convert_code)
+        return True
         
     def __repr__(self):
         return "<CStructOrUnionType %s %s%s>" % (self.name, self.cname,
@@ -899,12 +1018,17 @@ class CStructOrUnionType(CType):
     def attributes_known(self):
         return self.is_complete()
 
+    def can_be_complex(self):
+        # Does the struct consist of exactly two floats?
+        fields = self.scope.var_entries
+        return len(fields) == 2 and fields[0].type.is_float and fields[1].type.is_float
+
 
 class CEnumType(CType):
     #  name           string
     #  cname          string or None
     #  typedef_flag   boolean
-    
+
     is_enum = 1
     signed = 1
     rank = -1 # Ranks below any integer type
@@ -948,34 +1072,23 @@ class CStringType:
 
     def literal_code(self, value):
         assert isinstance(value, str)
-        return '"%s"' % Utils.escape_byte_string(value)
+        return '"%s"' % StringEncoding.escape_byte_string(value)
 
 
-class CUTF8StringType:
-    #  Mixin class for C unicode types.
-
-    is_string = 1
-    is_unicode = 1
-    
-    to_py_function = "PyUnicode_DecodeUTF8"
-    exception_value = "NULL"
-
-    def literal_code(self, value):
-        assert isinstance(value, str)
-        return '"%s"' % Utils.escape_byte_string(value)
-
-
-class CCharArrayType(CStringType, CArrayType):
+class CUTF8CharArrayType(CStringType, CArrayType):
     #  C 'char []' type.
     
     parsetuple_format = "s"
     pymemberdef_typecode = "T_STRING_INPLACE"
+    is_unicode = 1
+    
+    to_py_function = "PyUnicode_DecodeUTF8"
+    exception_value = "NULL"
     
     def __init__(self, size):
         CArrayType.__init__(self, c_char_type, size)
 
-
-class CUTF8CharArrayType(CUTF8StringType, CArrayType):
+class CCharArrayType(CStringType, CArrayType):
     #  C 'char []' type.
     
     parsetuple_format = "s"
@@ -995,6 +1108,17 @@ class CCharPtrType(CStringType, CPtrType):
         CPtrType.__init__(self, c_char_type)
 
 
+class UnspecifiedType(PyrexType):
+    # Used as a placeholder until the type can be determined.
+        
+    def declaration_code(self, entity_code, 
+            for_display = 0, dll_linkage = None, pyrex = 0):
+        return "<unspecified>"
+    
+    def same_as_resolved_type(self, other_type):
+        return False
+        
+
 class ErrorType(PyrexType):
     # Used to prevent propagation of error messages.
     
@@ -1004,6 +1128,9 @@ class ErrorType(PyrexType):
     to_py_function = "dummy"
     from_py_function = "dummy"
     
+    def create_convert_utility_code(self, env):
+        return True
+    
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
         return "<error>"
@@ -1059,8 +1186,8 @@ c_longdouble_type =  CFloatType(8)
 
 c_null_ptr_type =     CNullPtrType(c_void_type)
 c_char_array_type =   CCharArrayType(None)
-c_utf8_char_array_type =   CUTF8CharArrayType(None)
 c_char_ptr_type =     CCharPtrType()
+c_utf8_char_array_type = CUTF8CharArrayType(None)
 c_char_ptr_ptr_type = CPtrType(c_char_ptr_type)
 c_py_ssize_t_ptr_type =  CPtrType(c_py_ssize_t_type)
 c_int_ptr_type =      CPtrType(c_int_type)
@@ -1070,9 +1197,11 @@ c_returncode_type =   CIntType(2, 1, "T_INT", is_returncode = 1)
 c_anon_enum_type =    CAnonEnumType(-1, 1)
 
 # the Py_buffer type is defined in Builtin.py
-c_py_buffer_ptr_type = CPtrType(CStructOrUnionType("Py_buffer", "struct", None, 1, "Py_buffer"))
+c_py_buffer_type = CStructOrUnionType("Py_buffer", "struct", None, 1, "Py_buffer")
+c_py_buffer_ptr_type = CPtrType(c_py_buffer_type)
 
 error_type =    ErrorType()
+unspecified_type = UnspecifiedType()
 
 lowest_float_rank = 6
 
@@ -1099,6 +1228,10 @@ sign_and_rank_to_type = {
     (1, 6): c_float_type, 
     (1, 7): c_double_type,
     (1, 8): c_longdouble_type,
+# In case we're mixing unsigned ints and floats...
+    (0, 6): c_float_type, 
+    (0, 7): c_double_type,
+    (0, 8): c_longdouble_type,
 }
 
 modifiers_and_name_to_type = {
@@ -1126,28 +1259,55 @@ modifiers_and_name_to_type = {
     (2, 1, "int"): c_slong_type,
     (2, 2, "int"): c_slonglong_type,
     (2, 0, "Py_ssize_t"): c_py_ssize_t_type,
+    
+    (1, 0, "long"): c_long_type,
+    (1, 0, "short"): c_short_type,
+    (1, 0, "longlong"): c_longlong_type,
+    (1, 0, "bint"): c_bint_type,
 }
 
 def widest_numeric_type(type1, type2):
     # Given two numeric types, return the narrowest type
     # encompassing both of them.
     if type1.is_enum and type2.is_enum:
-        widest_type = c_int_type
-    elif type2.rank > type1.rank:
-        widest_type = type2
+        return c_int_type
+    elif type1 is type2:
+        return type1
+    elif (type1.signed and type2.signed) or (not type1.signed and not type2.signed):
+        if type2.rank > type1.rank:
+            return type2
+        else:
+            return type1
     else:
-        widest_type = type1
+        return sign_and_rank_to_type[min(type1.signed, type2.signed), max(type1.rank, type2.rank)]
     return widest_type
 
 def simple_c_type(signed, longness, name):
     # Find type descriptor for simple type given name and modifiers.
     # Returns None if arguments don't make sense.
     return modifiers_and_name_to_type.get((signed, longness, name))
+    
+def parse_basic_type(name):
+    base = None
+    if name.startswith('p_'):
+        base = parse_basic_type(name[2:])
+    elif name.startswith('p'):
+        base = parse_basic_type(name[1:])
+    elif name.endswith('*'):
+        base = parse_basic_type(name[:-1])
+    if base:
+        return CPtrType(base)
+    elif name.startswith('u'):
+        return simple_c_type(0, 0, name[1:])
+    else:
+        return simple_c_type(1, 0, name)
 
 def c_array_type(base_type, size):
     # Construct a C array type.
     if base_type is c_char_type:
         return CCharArrayType(size)
+    elif base_type is error_type:
+        return error_type
     else:
         return CArrayType(base_type, size)
 
@@ -1155,8 +1315,23 @@ def c_ptr_type(base_type):
     # Construct a C pointer type.
     if base_type is c_char_type:
         return c_char_ptr_type
+    elif base_type is error_type:
+        return error_type
     else:
         return CPtrType(base_type)
+        
+def Node_to_type(node, env):
+    from ExprNodes import NameNode, AttributeNode, StringNode, error
+    if isinstance(node, StringNode):
+        node = NameNode(node.pos, name=node.value)
+    if isinstance(node, NameNode) and node.name in rank_to_type_name:
+        return simple_c_type(1, 0, node.name)
+    elif isinstance(node, (AttributeNode, NameNode)):
+        node.analyze_types(env)
+        if not node.entry.is_type:
+            pass
+    else:
+        error(node.pos, "Bad type")
 
 def public_decl(base, dll_linkage):
     if dll_linkage:
