@@ -40,7 +40,7 @@ def file_newer_than(path, time):
     ftime = modification_time(path)
     return ftime > time
 
-# support for source file encoding detection and unicode decoding
+# support for source file encoding detection
 
 def encode_filename(filename):
     if isinstance(filename, unicode):
@@ -77,41 +77,39 @@ def open_source_file(source_filename, mode="rU"):
     encoding = detect_file_encoding(source_filename)
     return codecs.open(source_filename, mode=mode, encoding=encoding)
 
-class EncodedString(unicode):
-    # unicode string subclass to keep track of the original encoding.
-    # 'encoding' is None for unicode strings and the source encoding
-    # otherwise
-    encoding = None
-
-    def byteencode(self):
-        assert self.encoding is not None
-        return self.encode(self.encoding)
-
-    def utf8encode(self):
-        assert self.encoding is None
-        return self.encode("UTF-8")
-
-    def is_unicode(self):
-        return self.encoding is None
-    is_unicode = property(is_unicode)
-
-#    def __eq__(self, other):
-#        return unicode.__eq__(self, other) and \
-#            getattr(other, 'encoding', '') == self.encoding
-
-def escape_byte_string(s):
-    s = s.replace('\0', r'\x00')
-    try:
-        s.decode("ASCII")
-        return s
-    except UnicodeDecodeError:
-        pass
-    l = []
-    append = l.append
-    for c in s:
-        o = ord(c)
-        if o >= 128:
-            append('\\x%X' % o)
+def long_literal(value):
+    if isinstance(value, basestring):
+        if len(value) < 2:
+            value = int(value)
+        elif value[0] == 0:
+            value = int(value, 8)
+        elif value[1] in 'xX':
+            value = int(value[2:], 16)
         else:
-            append(c)
-    return ''.join(l)
+            value = int(value)
+    return not -2**31 <= value < 2**31
+
+# a simple class that simplifies the usage of utility code
+
+class UtilityCode(object):
+    def __init__(self, proto=None, impl=None, init=None, cleanup=None):
+        self.proto = proto
+        self.impl = impl
+        self.init = init
+        self.cleanup = cleanup
+
+    def write_init_code(self, writer, pos):
+        if not self.init:
+            return
+        if callable(self.init):
+            self.init(writer, pos)
+        else:
+            writer.put(self.init)
+
+    def write_cleanup_code(self, writer, pos):
+        if not self.cleanup:
+            return
+        if callable(self.cleanup):
+            self.cleanup(writer, pos)
+        else:
+            writer.put(self.cleanup)
